#include "synth.h"
#include "ac1.h"
#include <string>
#include <vector>
#include <climits>
#include <cstdlib>
#include <ctime>
#include <cmath>
#include <list>
#include <map>
#include <iomanip>
using namespace std;
#include "cmdline.h"

extern int CBITS;
extern int COSTS[];

/****************************************************************************/
int MIN_SHIFT = 0;
int MAX_SHIFT = 20;
int VERBOSE = 0;
coeff_t MAX_NUM    = 1 << MAX_SHIFT;
coeff_t MAX_PRINT  = MAX_NUM >> 1;
coeff_t MAX_GEN    = MAX_PRINT;
coeff_t MAX_NPRINT = 5;

typedef enum { MODE_SINGLE, MODE_RANDOM, MODE_EXHAUST } synth_mode_t;
synth_mode_t MODE = MODE_SINGLE;

int NTESTS;    /* for MODE_RANDOM */
int TEST_SIZE; /* for MODE_RANDOM */

int HYB_CUT = 27;

typedef enum { EST_OPT_TABLE, /* Est(c) = optimal cost of c, i.e. complexity */
       EST_VALUE,     /* Est(c) = ceil(log_2(c)) */
       EST_HYBRID,    /* Est(c) = ceil(log_2(c)) if c > 2^27, csd_bits(c) else */
       EST_CSD        /* Est(c) = # non-zero CSD bits of c */
} est_mode_t;
est_mode_t EST_MODE = EST_CSD;

typedef enum { ACM_OUT_CODE, ACM_OUT_GAP, ACM_OUT_DOT, ACM_OUT_DOT_CODE } acm_out_t;
acm_out_t ACM_OUTPUT = ACM_OUT_CODE;
bool ACM_HEADER = true;
string ACM_DOT_FILE = "./acm.dot";

int DIST_TEST = 3; /* either 2 or 3 */
bool NO_WEIGHT     = false; /* NO_WEIGHT sets the weight function W(dist)=1 */
bool GEN_UNIQUE    = false;
bool GEN_CODE      = false;
bool GEN_ADDERS    = false;
bool GEN_DAG       = false;
bool GEN_DOT       = false;
bool ALLOCATE_REGS = false;
bool PRINT_TARGETS = false;
bool USE_BOUND     = true;
bool USE_SUCCS2    = false;
bool USE_AC1       = false;
bool IMPROVE_AC1   = false;
bool USE_C1_ALGO   = false;
bool MAX_BENEFIT = false;
bool BHM = false;
bool OUTPUT_GAP = false;

bool SPLIT_2 = false;
int SPLIT_THRESH = 6;
#define if_verbose(x) if(VERBOSE >=x)

/************** Solver data ****************/
cfvec_t SIZES;
cfvec_t SPEC_ALL_TARGETS; /* all specified targets (including duplicates) */
cfset_t SPEC_TARGETS;     /* duplicates/evens/negatives are omitted here */
cfset_t TARGETS;          /* not yet synthesized coefficients */
cfset_t WORKLIST; 
cfset_t TRIGGERS; bool TRIG = false;
cfset_t GIVEN, SUCCS2;
full_addmap_t READY; /* synthesized coefficients */
full_addmap_t SUCCS; /* successors of READY */
distmap_t DISTS;     /* distance cache */
cfset_t COST_0, COST_1, COST_2;
int CURCOST, ITER, HEUR;

coeff_t cumulative_benefit(coeff_t c);

long SAVE_DIST(coeff_t succ, coeff_t t, long dist, bool print = false) {
    DISTS[cfpair_t(succ,t)] = dist;
    if(print) { cerr << "dist " << succ << " " << t << " " << dist << endl; }
    return dist;
}

coeff_t (*BENEFIT)(coeff_t) = cumulative_benefit;
coeff_t BEST_BENEFIT        = COEFF_MAX;

/**************************** Code generation *******************************/
static oplist_t *OPS;
static regmap_t *REGS;
static reg_t   (*TMPREG)();
adag *ADAG = NULL;

/**************************** **** *******************************/
/**************************** ALGO *******************************/
/**************************** **** *******************************/
inline int cost_estimate(coeff_t c);
void gen_code(coeff_t c, const adder &ad);
void synthesize(coeff_t c, adder ad);
adder *get_adder(coeff_t c);
void ac1_decompose_cheapest();
void print_set(char *nam, full_addmap_t &s, int max_nprint = MAX_NPRINT, ostream &ostr = cerr);
void print_set(char *nam, cfset_t &s, int max_nprint = MAX_NPRINT, ostream &ostr = cerr);


inline int lookup_dist(coeff_t succ, coeff_t t) {
    distmap_t::const_iterator d = DISTS.find(cfpair_t(succ,t));
    if(d!=DISTS.end()) return (*d).second; 
    else return -1;
}

inline int lookup_dist(coeff_t t) {
    distmap_t::const_iterator d = DISTS.find(cfpair_t(1,t));
    if(d!=DISTS.end()) return (*d).second;
    else return -1;
}

coeff_t _add_dist(coeff_t succ, coeff_t t) { 
    cfset_t s; int mincost = INT_MAX;

    compute_fast_sumdiffs(s, succ, t);
    FORALL(s, i) { 
	if(CONTAINS(SUCCS, *i)) return 2;
	coeff_t c = 1+cost_estimate(*i); 
	if(c < mincost) {
	    if(c==2) return 2;
	    else mincost = c;
	}
    }
    return mincost;
}

int estimate_add_dist(coeff_t succ, coeff_t t) {
    cfset_t s,ss; 
    coeff_t mincost = INT_MAX;

    compute_fast_sumdiffs(s, succ, t);
    FORALL(s, i) { 
	coeff_t c = 1 + cost_estimate(*i); /* (*i) should not be a cost-0 number (=1), */
	if(c < mincost) {                  /* it is caught by exact distance tests */
	    mincost = c;
	    if(mincost==DIST_TEST) return SAVE_DIST(succ, t, DIST_TEST);
	}
        if(mincost > 3 && CONTAINS(SUCCS2, *i)) { 
	    mincost = 3; 
	    if(mincost==DIST_TEST) return SAVE_DIST(succ, t, DIST_TEST);
	}
    }

    /** mincost >= 3 */
    if(DIST_TEST==3 && mincost > 3) {
	cfset_t prods, _succ; ADD(_succ, succ);
	gen_mul_coeffs(COST_1, _succ, prods); // subtract_set(prods, COST_1); REMOVE(prods, succ);
	FORALL(prods, p) {
	    coeff_t cc = 1 + _add_dist(*p, t); /* forward */
	    if(cc<mincost) { mincost = cc; if(mincost==3) return SAVE_DIST(succ, t, 3); }
	}

	FORALL(COST_1, j) {
	    if(t % (*j) != 0) continue; /* backward */
	    coeff_t cc = 1 + _add_dist(succ, t / (*j));
	    if(cc<mincost) { mincost = cc; if(mincost==3) return SAVE_DIST(succ, t, 3); }
	}
    }
    return SAVE_DIST(succ, t, mincost);
}

coeff_t maxdepth(const full_addmap_t &coeffs);

inline coeff_t W(coeff_t dist) {
    //if(dist<3) dist=3;
    if(NO_WEIGHT || dist > 11) return 1;
    else return (coeff_t)(pow(13.0, 11.0-dist));
}

coeff_t cumulative_benefit(coeff_t succ) {
    coeff_t total_benefit = 0;
    coeff_t max_benefit = 0;
    coeff_t b, bonus;
    FORALL(TARGETS, t) {
	if(succ == *t) return INT_MAX;
	else {
	    coeff_t R_t_dist = lookup_dist(*t);
	    coeff_t Rs_t_dist = lookup_dist(succ, *t);
	    assert(R_t_dist!=-1);

	    if ( Rs_t_dist == -1 ) {
		/* do not estimate if target is within exact distance test range */
		if ( R_t_dist <= DIST_TEST ) Rs_t_dist = R_t_dist;
	        /* estimate */
		else Rs_t_dist = estimate_add_dist(succ, *t);
	    }

	    if(Rs_t_dist < R_t_dist) b = (R_t_dist - Rs_t_dist) * W(Rs_t_dist);
	    else b = 0;

	    total_benefit += b;
	    if(b > max_benefit) max_benefit = b;
	    if(MAX_BENEFIT && b > 0 && Rs_t_dist == 1) return BEST_BENEFIT;
	}
    }

    return MAX_BENEFIT ? max_benefit : total_benefit;
}

#define Astar compute_fast_sumdiffs
bool intersect_save_dist(cfset_t &A, full_addmap_t &S, coeff_t t, long d) {
    bool found_intersection = false;
    FORALL(A, a) {
	if(CONTAINS(SUCCS, *a)) {
	    SAVE_DIST( 1, t, d);
	    SAVE_DIST(*a, t, d - 1);
	    found_intersection = true;
	}
    }
    return found_intersection;
}

void init_target_dists() {
    FORALL(TARGETS, tt) 
	SAVE_DIST(1,*tt,1000);
}

void target_dists_exact() {
    FORALL(TARGETS, tt) {
	coeff_t t = (*tt);
	cfset_t t_div_c1, t_div_c2;
	long dist = lookup_dist(1,t);
	coeff_t succ;

	/* Distance 2 Exact Tests -------------------------------*/
	if(dist == 1) continue;

	/* t_div_c1 ^ S */ // dist 2, case 2 
	set_ratio(t_div_c1, t, COST_1);
	if(intersect_save_dist(t_div_c1, SUCCS, t, 2)) dist=2;

        FORALL_FAM(READY, rr) {
	    coeff_t r = (*rr).first;
	    cfset_t rt;

	    /* rt ^ S */  // dist 2 case 1 
	    Astar(rt, r, t);
	    if(intersect_save_dist(rt, SUCCS, t, 2)) dist=2;
	}

	/* Distance 3 Exact Tests -------------------------------*/
	if(dist < 3 || DIST_TEST < 3) continue;

	/* t_div_c2 ^ S */ // dist 3, cases 1,2 
	set_ratio(t_div_c2, t, COST_2);
	if(intersect_save_dist(t_div_c2, SUCCS, t, 3)) dist=3;

        FORALL_FAM(READY, rr) {
	    coeff_t r = (*rr).first;
	    cfset_t rt, rt2, rt_div_c1;
	    Astar(rt, r, t);

	    /* rt2 ^ S */ // dist 3 case 3
	    FORALL(t_div_c1, z) Astar(rt2, r, *z);
	    if(intersect_save_dist(rt2, SUCCS, t, 3)) dist=3;
	
	    /* rt_div_c1 ^ S */ //dist 3 case 4
	    set_ratio(rt_div_c1, rt, COST_1);
	    if(intersect_save_dist(rt_div_c1, SUCCS, t, 3)) dist=3;
	}

	FORALL_FAM(SUCCS, ss) {
	    coeff_t s = (*ss).first;
	    cfset_t st;
	    Astar(st, s, t);
	    if(intersect_save_dist(st, SUCCS, t, 3)) dist=3;
	}
    }

}
	    
/* update dist to all targets once a successor is selected to be synthesized 
   in the heuristic part */
void update_target_dists(coeff_t succ) {
    FORALL(TARGETS, tt) {
	coeff_t t = *tt;
	long dist = lookup_dist(succ, t);
	if ( dist > 0 ) {
	  long orig_dist = lookup_dist(1,t);
	  if(orig_dist > dist)
	      SAVE_DIST(1,t,dist);
	}
    }
}

void heur_synth_best() {
    coeff_t best_benefit = 0;
    coeff_t best_coeff = 0;
    adder   best_adder;

    FORALL_FAM(SUCCS, s) {
	coeff_t c = (*s).first;
	coeff_t benefit = BENEFIT(c);
	if_verbose(3) {
	    if(ITER==0 && CONTAINS(COST_1, c)) cerr << endl << c << "-" << benefit << ", "; }

	if(benefit == BEST_BENEFIT) {
	    best_coeff = c;
	    best_adder = (*s).second;
	    break;
	}
	else if(benefit > best_benefit) {
	    best_benefit = benefit;
	    best_coeff = c;
	    best_adder = (*s).second;
	}
	else if(benefit == best_benefit && (*s).second.depth < best_adder.depth) {
	    best_coeff = c;
	    best_adder = (*s).second;
	}
    }

    ASSERT(best_coeff != 0);
    update_target_dists(best_coeff);
    synthesize(best_coeff, best_adder);
}

void synthesize(coeff_t c, adder ad) {
    if(! (CONTAINS(READY, ad.c1) && CONTAINS(READY, ad.c2))) {
	cerr << "offending adder:";
	ad.output(cerr); cerr << " actual=" << ad.res << endl;
	assert(CONTAINS(READY, ad.c1) && CONTAINS(READY, ad.c2));
    }
    if(ad.depth==-1)
	ad.depth = 1 + max(READY[ad.c1].depth, READY[ad.c2].depth);
    REMOVE(TARGETS, c);
    REMOVE(SUCCS, c);
    REMOVE(SUCCS2, c);
    READY[c] = ad;
    ADD(WORKLIST, c);

    if(GEN_CODE)   gen_code(c, ad);
    if(GEN_ADDERS) { ad.output(cout); cout << endl; }
    if(GEN_DAG) { 
	reg_t dest;
	if(REGS->find(c) != REGS->end()) dest = (*REGS)[c];
	else dest = TMPREG();
	adag_node * n = ADAG->add_adder(&ad, dest, TMPREG, get_adder); 
	if_verbose(1) { 
	    cerr << n->depth << " " << setw(10) << n->compute_coeff() << " ";
	    n->output(cerr);
	}
    }
    if_verbose(2) cerr << c << " ";
    ++CURCOST;
}

void compute_succs2(coeff_t newsucc) {
    REMOVE(SUCCS2, newsucc);
    set_product(SUCCS2, newsucc, READY);
    compute_fast_sumdiffs(SUCCS2, newsucc, newsucc);
}

void compute_succs() {
    addmap_t newsuccs;
    cfset_t work = WORKLIST;
    full_addmap_t targets;
    WORKLIST.clear();

    if(USE_SUCCS2) 
	set_product(SUCCS2, SUCCS, work);

    FORALL_FAM(READY, rr) {
	coeff_t r = (*rr).first;
	int rdepth = (*rr).second.depth;

	FORALL(work, w) {
	    int wdepth = READY[*w].depth;
	    int depth = 1 + max(rdepth, wdepth);
	    newsuccs.clear();
	    compute_all_sumdiffs(newsuccs, r, *w);

	    FORALL_AM(newsuccs, s) {
		coeff_t c = (*s).first;
		sp_t sp = (*s).second;

		if(CONTAINS(TARGETS, c)) {
		    if(!CONTAINS(targets, c) || depth < targets[c].depth)
			targets[c] = adder(true, r, *w, sp, depth);
		}
		else if(! CONTAINS(READY, c)) {
		    bool is_new = ! CONTAINS(SUCCS, c);
		    if(USE_SUCCS2 && is_new) 
			compute_succs2(c);
		    if(is_new || depth < SUCCS[c].depth) {
			SUCCS[c] = adder(false, r, *w, sp, depth);
		    }
		}
	    }
	}
    }
    FORALL_FAM(targets, c) synthesize((*c).first, (*c).second);
}

void dump_state(ostream &os) {
    os << ITER << " | " 
       << "curcost " << CURCOST << "/" << maxdepth(READY) << " | " 
       << "succs " << SUCCS.size() << " " << SUCCS2.size() << " | ";
    FORALL(TARGETS, t) 
	os << lookup_dist(*t) << " ";
    os << "|| ";
    if_verbose(3) print_set("targets:", TARGETS, 0, os);
}

void _solve() {
    coeff_t best;
    FORALL_FAM(READY, r) ADD(WORKLIST, (*r).first);
    CURCOST = ITER = HEUR = 0;
    init_target_dists();
	
    while(!TARGETS.empty()) {
	while(!WORKLIST.empty()) compute_succs(); 

	target_dists_exact();
	if_verbose(2) { cerr << endl; dump_state(cerr); }

	if(!TARGETS.empty())     
	    if(!USE_AC1) heur_synth_best(); 
	    else ac1_decompose_cheapest();

	++ITER; 
    }
}

/*****************************************************************************/
#include "bhm.cpp"

void create_problem(cfset_t &coeffs, reg_t);

void solve(oplist_t *l, regmap_t *regs, reg_t (*tmpreg)()) {
    OPS = l; REGS = regs; TMPREG = tmpreg;
    if(GEN_CODE || GEN_DAG) {
	if(REGS==NULL) REGS = new regmap_t;
	ASSERT(TMPREG!=NULL);
    }

    if(USE_C1_ALGO) dempster_c1_solve();
    else if(BHM) bhm_solve();
    else _solve();

    if(GEN_DAG) { 
	ADAG->to_multi();
	if(GEN_DOT) { cout << "digraph z{\n"; ADAG->output_dot(cout); cout << "}\n"; }
	if_verbose(1) { 
	    cerr << "--- reduced graph ---\n"; 
	    ADAG->output_code(cerr); 
	}
	if(ALLOCATE_REGS) {
	    ADAG->compute_liveness(true);
	    ADAG->allocate_regs();

	    if_verbose(1) { 
		cerr << " --- liveness ---\n"; 
		ADAG->output_liveness(cerr); 
		cerr << "// minregs = " << ADAG->min_regs() << endl;
		cerr << "// allocated = " << ADAG->nregs << endl;
	    }
	}
    }
    if(regs==NULL && REGS!=NULL) delete REGS;
}

void create_problem(cfset_t &coeffs, reg_t input_reg) {
    SUCCS.clear();   SUCCS2.clear();  TARGETS.clear();
    READY.clear();   WORKLIST.clear(); 
    DISTS.clear();   TRIGGERS.clear();
    if(GEN_DAG) { if(ADAG!=NULL) delete ADAG; ADAG = new adag(input_reg); }

    coeff_t max = 1, q;
    FORALL(coeffs, c) {
	coeff_t tt = fundamental(cfabs(*c));

	if(SPLIT_2 && csd_bits(tt) >= SPLIT_THRESH) {
	    long csd = csd_bits(tt);
	    long ofs=0;
	    while(csd_bits(tt>>ofs) > 3*csd/4) ++ofs;
	    ADD(TARGETS, (q=fundamental(tt>>ofs))); if(q>max) max=q;
	    ADD(TARGETS, (q=fundamental(tt & ((1<<ofs)-1)))); if(q>max) max=q;
	}
	else {
	    ADD(TARGETS, tt); if(tt>max) max=tt;
	}
    }
    set_algo_bits(clog2(max)); 
    //if_verbose(1) { cerr << "Effective bits = " << clog2(max) << endl; }
    FORALL(GIVEN, g) {
	coeff_t gg = fundamental(cfabs(*g));
	READY[gg] = adder(gg);
	REMOVE(TARGETS, gg);
    }
    REMOVE(TARGETS, 1);
    SPEC_TARGETS = TARGETS;
}

coeff_t rand_coeff(coeff_t max) {
    coeff_t res = 0;
    /* we will generate 8 bits at a time */
    assert(RAND_MAX >= (1<<8)-1);
    for(int bytes=0; bytes < sizeof(coeff_t); ++bytes) {
	res = (res<<8) + rand() % (1<<8);
	//printf("%lld \n", res);
    }
    return 1 + (cfabs(res) % max);
}

void create_random_problem(int size) {
    cfset_t problem;
    for(int i = 0; i < size; ++i) {
	coeff_t c;
	if(GEN_UNIQUE) {
	    do c = rand_coeff(MAX_GEN) | 1; /* odd numbers only */
	    while ( c==1 || CONTAINS(problem, c) );
	    ASSERT(fundamental(c)==c && c!=1);
	}
	else c = fundamental(rand_coeff(MAX_GEN));

	ADD(problem, c);
    }
    create_problem(problem);
}

void init_synth(int argc, char **argv) {
    ADD(GIVEN, 1);
    parse_cmdline(cmditer_new(argc, argv));
    ADD(COST_0, 1);
    gen_add_coeffs(COST_0, COST_0, COST_1);
    gen_add_coeffs(COST_0, COST_1, COST_2);
    gen_mul_coeffs(COST_1, COST_1, COST_2);

    subtract_set(COST_1, COST_0);
    subtract_set(COST_2, COST_0);
    subtract_set(COST_2, COST_1);
}
